﻿using System;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;

/*
 *         Progetto: Progetto "Trasmissioni multicast"
 * 
 *           Autore: Roberto Fuligni
 *  Ultima modifica: 13/06/2016
 *  
 *      Descrizione: Funzioni di utilità per l'invio e la ricezione 
 *                   di messaggi multicast.
 *                   
 */

namespace LibreriaMulticast
{
    public static class UtilityMulticast
    {
        // Determina l'indirizzo IPV4 dell'interfaccia di rete collegata alla LAN
        // assumendo che questa sia l'unica ad avere un gateway configurato.

        public static IPAddress RicercaIpHostLAN()
        {

            NetworkInterface[] interfaces = NetworkInterface.GetAllNetworkInterfaces();
            foreach (NetworkInterface adapter in interfaces)
            {
                IPInterfaceProperties prop = adapter.GetIPProperties();
                if (prop.GatewayAddresses.Count > 0)
                    foreach (var a in adapter.GetIPProperties().UnicastAddresses)
                    {
                        if (a.Address.AddressFamily == AddressFamily.InterNetwork)
                            return a.Address;
                    }
            }

            throw new Exception("Nessuna interfaccia con gateway configurato.");
        }

        private static string MessaggioTest(IPEndPoint mittente, IPEndPoint destinatario)
        {
            return String.Format("Test UDP del {0:dd/MM/yyyy H:mm:ss.fff} da {1} a {2}", 
                DateTime.Now, 
                mittente, 
                destinatario);
        }

        // Invia un messaggio in multicast all'endpoint specificato (es. 244.4.4.4:2016)
        // attraverso una specifica interfaccia di rete locale
        public static string InviaMessaggio(IPEndPoint epMulticast, IPAddress ipLocale)
        {
            string msg;

            using (var sck = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp))
            {
                MulticastOption opzione = new MulticastOption(epMulticast.Address, ipLocale);

                sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.AddMembership, opzione);
                sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.MulticastTimeToLive, 1);

                sck.Bind(new IPEndPoint(ipLocale, 0));  // Porta scelta automaticamente

                msg = MessaggioTest((IPEndPoint)sck.LocalEndPoint, epMulticast);
                Byte[] sendBytes = Encoding.UTF8.GetBytes(msg);

                sck.Connect(epMulticast);
                sck.Send(sendBytes);

                sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.DropMembership, opzione);

                return msg;
            }
        }

        // Avvia e restituisce un listener UDP in grado di ricevere pacchetti multicast
        // La gestione dei pacchetti ricevuti è affidata a un thread separato
        public static Socket AvviaListenerUDP(IPEndPoint epMulticast, IPAddress ipLocale, Action<string, IPEndPoint, IPEndPoint> azione)
        {
            

            Socket sck = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);

            MulticastOption opzione = new MulticastOption(epMulticast.Address, ipLocale);

            sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.AddMembership, opzione);
            sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.MulticastTimeToLive, 1);
            sck.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);

            sck.Bind(new IPEndPoint(ipLocale, epMulticast.Port));

            Task.Run(() =>
            {
                bool finito = false;
                EndPoint epMittente = (EndPoint)new IPEndPoint(IPAddress.Any, 0);

                while (!finito)
                {
                    try
                    {
                        int letti;
                        byte[] buffer = new byte[2048];
                        IPPacketInformation info;
                        SocketFlags flags = new SocketFlags();
                        string msg;

                        letti = sck.ReceiveMessageFrom(buffer, 0, buffer.Length, ref flags, ref epMittente, out info);
                        msg = Encoding.UTF8.GetString(buffer, 0, letti);
                        IPEndPoint epDestinatario = (IPEndPoint)sck.LocalEndPoint;

                        azione(msg, (IPEndPoint) epMittente, new IPEndPoint(info.Address, epDestinatario.Port));
                    }
                    catch (Exception)
                    {
                        finito = true;
                    }
                }
            });

            return sck;

        }

        // Chiude un listener e conclude la l'attività del thread associato
        public static void ChiudiListenerUdp(Socket sck, IPEndPoint epMulticast, IPAddress ipLocale)
        {
            MulticastOption op = new MulticastOption(epMulticast.Address, ipLocale);
            sck.SetSocketOption(SocketOptionLevel.IP, SocketOptionName.DropMembership, op);
            sck.Dispose();  // Termina l'esecuzione del thread di elaborazione dei messaggi
        }
    }
}
